import os
from pathlib import Path as PathlibPath
from sys import path as sys_path

from click import Path, argument, group, option
from trilogy.dialect.enums import Dialects
from trilogy.parsing.render import Renderer

# handles development cases
nb_path = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
sys_path.insert(0, nb_path)

from trilogyt.scripts.dbt import dbt_wrapper, dbt_string_command_wrapper  # noqa
from trilogyt.scripts.native import (  # noqa
    native_wrapper,
    native_string_command_wrapper,
)
from trilogyt.scripts.dagster import (  # noqa
    dagster_wrapper,
    dagster_string_command_wrapper,
)

renderer = Renderer()


@group()
def main():
    """A CLI with three subcommands: dbt, dagster and trilogy"""
    pass


@main.command()
@argument("preql", type=Path())
@argument("output_path", type=Path(exists=True))
# @argument("write_path", type=Path(exists=True))
@argument("dialect", type=str)
@option("--run", is_flag=True, type=bool, default=False)
@option("--debug", type=bool, default=False)
def dbt(preql: str | Path, output_path: Path, dialect: str, debug: bool, run: bool):

    edialect = Dialects(dialect)
    preqlt: PathlibPath = PathlibPath(str(preql))
    if preqlt.exists():
        return dbt_wrapper(preqlt, PathlibPath(str(output_path)), edialect, debug, run)
    return dbt_string_command_wrapper(
        str(preql), PathlibPath(str(output_path)), edialect, debug, run
    )


@main.command()
@argument("preql", type=Path())
@argument("output_path", type=Path(exists=True))
# @argument("write_path", type=Path(exists=True))
@argument("dialect", type=str)
@option("--run", is_flag=True, type=bool, default=False)
@option("--debug", type=bool, default=False)
def trilogy(preql: str | Path, output_path: Path, dialect: str, debug: bool, run: bool):

    edialect = Dialects(dialect)
    preqlt: PathlibPath = PathlibPath(str(preql))
    if preqlt.exists():
        return native_wrapper(
            preqlt, PathlibPath(str(output_path)), edialect, debug, run
        )
    return native_string_command_wrapper(
        str(preql), PathlibPath(str(output_path)), edialect, debug, run
    )


@main.command()
@argument("preql", type=Path())
@argument("output_path", type=Path(exists=True))
# @argument("write_path", type=Path(exists=True))
@argument("dialect", type=str)
@option("--run", is_flag=True, type=bool, default=False)
@option("--debug", type=bool, default=False)
def dagster(preql: str | Path, output_path: Path, dialect: str, debug: bool, run: bool):

    edialect = Dialects(dialect)
    preqlt: PathlibPath = PathlibPath(str(preql))
    if preqlt.exists():
        return dagster_wrapper(
            preqlt, PathlibPath(str(output_path)), edialect, debug, run
        )
    return dagster_string_command_wrapper(
        str(preql), PathlibPath(str(output_path)), edialect, debug, run
    )


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from hashlib import sha256
from random import choice
from typing import List

from trilogy import Environment
from trilogy.constants import VIRTUAL_CONCEPT_PREFIX
from trilogy.core.enums import PurposeLineage
from trilogy.core.ergonomics import CTE_NAMES
from trilogy.core.models import (
    CTE,
    Address,
    Comparison,
    Concept,
    Conditional,
    Datasource,
    Grain,
    Parenthetical,
    PersistStatement,
    ProcessedQuery,
    ProcessedQueryPersist,
    ProcessedRawSQLStatement,
    ProcessedShowStatement,
    QueryDatasource,
    RowsetDerivationStatement,
    SelectItem,
    SelectStatement,
    UnionCTE,
    WhereClause,
)
from trilogy.dialect.base import BaseDialect


@dataclass
class AnalyzeResult:
    counts: dict[str, int]
    mapping: dict[str, CTE]


@dataclass
class ProcessLoopResult:
    new: List[PersistStatement]


def name_to_short_name(x: str):
    if x.startswith(VIRTUAL_CONCEPT_PREFIX):
        return "virtual"
    return x


# hash a
def hash_concepts(concepts: list[Concept]) -> str:
    return sha256("".join([x.address for x in concepts]).encode()).hexdigest()


def generate_datasource_name(select: SelectStatement, cte_name: str) -> str:
    """Generate a reasonable table name"""
    base = "ds" + "_".join(
        [name_to_short_name(x) for x in select.grain.components]
        + [hash_concepts(select.output_components)]
    )
    if select.where_clause:
        # TODO: needs to include the filter predicates
        base = (
            base
            + "_filtered_on_"
            + "_".join(
                [
                    name_to_short_name(x.name)
                    for x in select.where_clause.conditional.concept_arguments
                ]
            )
        )
    return base


def cte_to_persist(input: CTE, name: str, generator: BaseDialect) -> PersistStatement:

    # a rowset will have the CTE ported over and we can assume the select statement doesn't need
    # further filtering
    rowset = any([x.derivation == PurposeLineage.ROWSET for x in input.output_columns])
    if rowset:
        select = SelectStatement(
            selection=[SelectItem(content=x) for x in input.output_columns],
        )
    else:
        select = SelectStatement(
            selection=[SelectItem(content=x) for x in input.output_columns],
            where_clause=(
                WhereClause(conditional=input.condition) if input.condition else None
            ),
        )
    datasource = select.to_datasource(
        namespace="default",
        name=generate_datasource_name(select, name),
        address=Address(location=generate_datasource_name(select, name)),
    )
    persist = PersistStatement(datasource=datasource, select=select)
    return persist


def remap_dictionary(d: dict, remap: list[str]) -> tuple[dict, dict]:
    new: dict[str, str] = {}
    mapping: dict[str, CTE] = {}
    for k, v in d.items():
        alias = None
        while alias is None or alias in new:
            alias = choice(remap)
        new[alias] = v
        mapping[k] = alias
    return new, mapping


def fingerprint_concept(concept: Concept) -> str:
    return concept.address


def fingerprint_grain(grain: Grain) -> str:
    return "-".join([x for x in grain.components])


def fingerprint_source(source: QueryDatasource | Datasource) -> str:
    if isinstance(source, Datasource):
        return str(source.address)
    return "-".join([fingerprint_source(s) for s in source.datasources]) + str(
        source.limit
    )


def fingerprint_filter(filter: Conditional | Comparison | Parenthetical) -> str:
    return str(filter)


def fingerprint_cte(cte: CTE | UnionCTE, select_condition) -> str:
    if isinstance(cte, UnionCTE):
        return "-".join(
            [fingerprint_cte(x, select_condition) for x in cte.internal_ctes]
        )
    else:
        base = (
            fingerprint_source(cte.source)
            + "-".join([fingerprint_cte(x, select_condition) for x in cte.parent_ctes])
            + fingerprint_grain(cte.grain)
        )

    if cte.condition:
        base += fingerprint_filter(cte.condition)
    if select_condition:
        base += fingerprint_filter(select_condition)
    return "ds" + base


def process_raw(
    inputs: List[RowsetDerivationStatement | SelectStatement | PersistStatement],
    env: Environment,
    generator: BaseDialect,
    threshold: int = 2,
    inject: bool = False,
) -> tuple[
    list[RowsetDerivationStatement | SelectStatement | PersistStatement],
    list[PersistStatement],
]:
    complete = False
    outputs = []
    x = 0
    while not complete:
        x += 1
        print(f" loop {x}")
        parsed = generator.generate_queries(env, inputs)
        new = process_loop(parsed, env, generator=generator, threshold=threshold)
        if new.new:
            outputs += new.new
            # hardcoded until we figure out the right exit criteria
            complete = True
            if inject:
                insert_index = min(
                    [inputs.index(x) for x in inputs if isinstance(x, SelectStatement)]
                )
                # insert the new values before this statement
                inputs = inputs[:insert_index] + new.new + inputs[insert_index:]
        else:
            complete = True
    return inputs, outputs


def analyze(
    inputs: List[ProcessedQuery | ProcessedQueryPersist | ProcessedShowStatement],
    env: Environment,
):
    counts: dict[str, int] = {}
    lookup: dict[str, CTE] = {}
    valid = [
        c for c in inputs if isinstance(c, (ProcessedQuery, ProcessedQueryPersist))
    ]
    if not valid:
        return AnalyzeResult({}, {})
    for x in valid:
        for cte in x.ctes:
            # if the CTE is being generated as part of a larger query with a filter condition
            # and it itself doesn't have a filter condition; this might be post filtering
            # and irrelevant to global graph.
            # TODO: figure out if it's pre-filtering or post-filtering
            if x.where_clause and not cte.condition:
                continue
            fingerprint = fingerprint_cte(
                cte, x.where_clause.conditional if x.where_clause else None
            )
            counts[fingerprint] = counts.get(fingerprint, 0) + 1
            lookup[fingerprint] = lookup.get(fingerprint, cte) + cte
    for k, v in counts.items():
        print(
            f"CTE {k} outputs {[x.address for x in lookup[k].output_columns]} is used {v} times"
        )
    counts, mapping = remap_dictionary(counts, CTE_NAMES)
    # for k, v in counts.items():
    #     print(f"CTE {k} is used {v} times")
    final_map: dict[str, CTE] = {}
    for fingerprint, cte in lookup.items():
        final_map[mapping[fingerprint]] = cte
    for k2, v2 in final_map.items():
        print(f"CTE {k2} outputs {[x.address for x in v2.output_columns]}")
    return AnalyzeResult(counts, final_map)


def is_raw_source_cte(cte: CTE) -> bool:
    if not len(cte.source.datasources) == 1:
        return False
    if isinstance(cte.source.datasources[0], Datasource):
        source = cte.source.datasources[0]
        if isinstance(source.address, Address) and source.address.is_query:
            return False
        return True
    return False


def has_anon_concepts(cte: CTE) -> bool:
    if any([x.name.startswith(VIRTUAL_CONCEPT_PREFIX) for x in cte.output_columns]):
        return True
    return False


def reorder_ctes(
    input: list[CTE],
):
    import networkx as nx

    # Create a directed graph
    G: nx.DiGraph = nx.DiGraph()
    mapping: dict[str, CTE] = {}
    for cte in input:
        mapping[cte.name] = cte
        for parent in cte.parent_ctes:
            G.add_edge(parent.name, cte.name)
    # Perform topological sort (only works for DAGs)
    try:
        topological_order = list(nx.topological_sort(G))
        if not topological_order:
            return input
        return [mapping[x] for x in topological_order]
    except nx.NetworkXUnfeasible as e:
        print(
            "The graph is not a DAG (contains cycles) and cannot be topologically sorted."
        )
        raise e


def process_loop(
    inputs: List[
        ProcessedQuery
        | ProcessedQueryPersist
        | ProcessedShowStatement
        | ProcessedRawSQLStatement
    ],
    env: Environment,
    generator: BaseDialect,
    threshold: int = 2,
) -> ProcessLoopResult:
    analysis = analyze(
        [x for x in inputs if isinstance(x, (ProcessedQuery, ProcessedQueryPersist))],
        env,
    )
    new_persist = []
    for k, v in analysis.counts.items():
        print(f"CTE {k} is used {v} times")
        if v >= threshold:
            cte = analysis.mapping[k]
            # skip materializing materialized things
            if is_raw_source_cte(cte):
                continue
            if has_anon_concepts(cte):
                continue
            print(
                f"Creating new persist statement from cte {k} with output {[x.address for x in cte.output_columns]}"
            )
            new_persist.append(cte_to_persist(cte, k, generator=generator))

    return ProcessLoopResult(new=new_persist)



from hashlib import sha256

from trilogy.core.enums import Purpose
from trilogy.core.env_processor import generate_graph
from trilogy.core.functions import CurrentDatetime
from trilogy.core.models import Concept, DataType, Environment

from trilogyt.constants import TRILOGY_NAMESPACE
from trilogyt.enums import PreqltMetrics

ENVIRONMENT_CONCEPTS = [
    Concept(
        name=PreqltMetrics.CREATED_AT.value,
        namespace=TRILOGY_NAMESPACE,
        datatype=DataType.DATETIME,
        purpose=Purpose.CONSTANT,
        lineage=CurrentDatetime([]),
    )
]


def enrich_environment(env: Environment):
    if TRILOGY_NAMESPACE in env.imports:
        raise Exception(f"{TRILOGY_NAMESPACE} namespace already exists")

    for concept in ENVIRONMENT_CONCEPTS:
        env.add_concept(concept)
    return env


def fingerprint_environment(env: Environment) -> str:
    graph = generate_graph(env)
    edges = "-".join([str(x) for x in sorted(list(graph.edges))])
    return sha256(edges.encode()).hexdigest()



from trilogy.dialect.enums import Dialects  # noqa
from pathlib import Path as PathlibPath  # noqa
import os
from sys import path as sys_path
from trilogy import Environment, Executor
from trilogy.parser import parse_text
from trilogy.parsing.render import Renderer
from trilogy.utility import unique
from trilogy.core.models import (
    ImportStatement,
    PersistStatement,
    SelectStatement,
    ConceptDeclarationStatement,
    CopyStatement,
    HasUUID,
)
from dataclasses import dataclass
from trilogyt.core import ENVIRONMENT_CONCEPTS, fingerprint_environment

# handles development cases
nb_path = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
sys_path.insert(0, nb_path)

from trilogyt.constants import OPTIMIZATION_NAMESPACE, OPTIMIZATION_FILE  # noqa
from trilogyt.graph import process_raw  # noqa
from trilogyt.exceptions import OptimizationError  # noqa


@dataclass
class OptimizationInput:
    fingerprint: str
    environment: Environment
    statements: list


@dataclass
class OptimizationResult:
    path: PathlibPath
    datasource_path: PathlibPath
    new_import: ImportStatement
    fingerprint: str


renderer = Renderer()


def print_tabulate(q, tabulate):
    result = q.fetchall()
    print(tabulate(result, headers=q.keys(), tablefmt="psql"))


def optimize_multiple(
    base: PathlibPath,
    paths: list[PathlibPath],
    output_path: PathlibPath,
    dialect: Dialects,
) -> dict[PathlibPath, OptimizationResult]:

    optimize_env = Environment(working_path=base.stem, namespace="optimize")
    exec = Executor(
        dialect=dialect, engine=dialect.default_engine(), environment=optimize_env
    )

    env_to_statements: dict[str, OptimizationInput] = {}
    file_to_fingerprint = {}
    for path in paths:
        if path.name.startswith(OPTIMIZATION_FILE):
            continue
        with open(path) as f:
            local_env = Environment(
                working_path=path.parent,
            )
            try:
                new_env, statements = parse_text(f.read(), environment=local_env)
            except Exception as e:
                raise SyntaxError(f"Unable to parse {path} due to {e}")
            if not any(
                isinstance(
                    statement, (SelectStatement, PersistStatement, CopyStatement)
                )
                for statement in statements
            ):
                continue
            fingerprint = fingerprint_environment(new_env)
            file_to_fingerprint[path] = fingerprint
            if fingerprint in env_to_statements:
                opt: OptimizationInput = env_to_statements[fingerprint]
                opt.statements += statements
            else:
                env_to_statements[fingerprint] = OptimizationInput(
                    fingerprint=fingerprint, environment=new_env, statements=statements
                )

    # determine the new persists we need to create
    outputs = {}
    for k, v in env_to_statements.items():
        _, new_persists = process_raw(
            inject=False,
            inputs=v.statements,
            env=v.environment,
            generator=exec.generator,
            threshold=2,
        )

        concept_modifying_statements = unique(
            [x for x in v.statements if isinstance(x, HasUUID)], "uuid"
        )
        final = []
        # we should transform a persist into a select for optimization purposes
        for x in concept_modifying_statements:
            if isinstance(x, PersistStatement):
                final.append(x.select)
            else:
                final.append(x)
        # inject those
        output_file = output_path / f"{OPTIMIZATION_FILE}_{k}.preql"
        with open(output_file, "w") as f:
            for concept in ENVIRONMENT_CONCEPTS:
                f.write(
                    renderer.to_string(ConceptDeclarationStatement(concept=concept))
                    + "\n\n"
                )
            for cte in final:
                f.write(renderer.to_string(cte) + "\n\n")
            for x in new_persists:
                f.write(renderer.to_string(x) + "\n\n")
                # f.write(renderer.to_string(x.datasource) + "\n\n")

        datasource_file = output_path / f"{OPTIMIZATION_FILE}_{k}_datasources.preql"
        with open(datasource_file, "w") as f:
            f.write(f"import {OPTIMIZATION_FILE}_{k};\n\n")
            for x in new_persists:
                f.write(renderer.to_string(x.datasource) + "\n\n")
        outputs[k] = OptimizationResult(
            path=output_file,
            datasource_path=datasource_file,
            new_import=ImportStatement(
                alias=OPTIMIZATION_NAMESPACE,
                path=output_file,
            ),
            fingerprint=k,
        )
    return {k: outputs[v] for k, v in file_to_fingerprint.items()}



from pathlib import Path
from typing import Optional

from pydantic import BaseModel

from trilogyt.dagster.constants import SUFFIX


class DagsterConfig(BaseModel):
    root: Path
    build_tests: bool = True
    test_path: str = "tests"
    dagster_asset_path: str = "assets"
    namespace: Optional[str] = None

    def get_asset_path(self, key: str) -> Path:
        if self.namespace:
            output_path = (
                self.root / self.dagster_asset_path / self.namespace / f"{key}_{SUFFIX}"
            )
        else:
            output_path = self.root / self.dagster_asset_path / f"{key}_{SUFFIX}"
        return output_path

    @property
    def config_path(self) -> Path:
        if self.namespace:
            return self.root / self.dagster_asset_path / self.namespace / "schema.yml"
        return self.root / self.dagster_asset_path / "schema.yml"



from trilogy import Dialects, parse
from trilogy.core.enums import SourceType
from trilogy.core.models import CTE, Environment, LooseConceptList, QueryDatasource
from trilogy.dialect.duckdb import DuckDBDialect
from trilogy.parsing.render import Renderer

from trilogyt.graph import fingerprint_cte, process_raw


def test_fingerprint(test_environment: Environment):

    oid = test_environment.concepts["order_id"]
    oid_ds = test_environment.datasources["orders"]
    qds = QueryDatasource(
        output_concepts=[oid],
        input_concepts=[oid],
        datasources=[oid_ds],
        source_map={oid.address: {oid_ds}},
        grain=oid_ds.grain,
        source_type=SourceType.DIRECT_SELECT,
        joins=[],
    )
    test = CTE(
        name="test",
        source=qds,
        output_columns=[oid],
        source_map={oid.address: [qds.name]},
        grain=qds.grain,
    )
    a = fingerprint_cte(test, None)
    test.name = "test2"
    b = fingerprint_cte(test, None)
    assert a == b


def test_integration():
    env = Environment()

    env, parsed = parse(
        """
key int_array list<int>;

datasource int_source (
int_array:int_array
)
grain (int_array)
query '''
select [1,2,3,4] as int_array
''';
                   
auto split <- unnest(int_array);
                   
select split;

select split;

select split;      

""",
        environment=env,
    )
    exec = Dialects.DUCK_DB.default_executor(environment=env)
    dialect = DuckDBDialect()
    initial = dialect.generate_queries(env, parsed)
    assert len(initial) == 3
    consolidated, new = process_raw(
        parsed, env=env, generator=dialect, threshold=2, inject=True
    )

    # we should have the one consolidated CTE first
    assert len(consolidated) == 8
    renderer = Renderer()
    final = []
    for x in consolidated:
        final.append(renderer.to_string(x))
    reparsed = exec.parse_text("\n".join(final), persist=True)

    # we should have our new datasource
    assert len(env.datasources) == 3
    env = exec.environment
    split = env.concepts["split"]
    instance = [
        x for x in list(env.datasources.values()) if split.address in x.output_concepts
    ][0]
    assert split.address in [x for x in env.materialized_concepts]
    assert "local.split" in [x for x in env.materialized_concepts]
    materialized_lcl = LooseConceptList(
        concepts=[
            x
            for x in reparsed[-1].output_columns
            if x.address in [z for z in env.materialized_concepts]
        ]
    )
    assert materialized_lcl.addresses == {"local.split"}
    final = reparsed[-1]
    # check that our queries use the new datasource
    assert final.ctes[0].source.datasources[0] == instance, final.ctes[
        0
    ].source.datasources[0]



from trilogy.constants import CONFIG

from trilogyt.core import enrich_environment
from trilogyt.dagster.config import DagsterConfig
from trilogyt.dagster.generate import generate_model as generate_dagster_model
from trilogyt.dbt.config import DBTConfig
from trilogyt.dbt.generate import generate_model as generate_dbt_model

# Do not use parameters when compiling SQL
# TODO: make this local to environments?
# likely requires trilogy feature request
CONFIG.rendering.parameters = False

__version__ = "0.0.13"

__all__ = [
    "enrich_environment",
    "generate_dbt_model",
    "DBTConfig",
    "generate_dagster_model",
    "DagsterConfig",
]



import os
from collections import Counter, defaultdict
from dataclasses import dataclass
from pathlib import Path

from jinja2 import Template
from trilogy import Environment, Executor
from trilogy.core.models import (
    Datasource,
    PersistStatement,
    ProcessedQueryPersist,
    UnionCTE,
)
from trilogy.dialect.enums import Dialects

from trilogyt.constants import logger
from trilogyt.core import enrich_environment
from trilogyt.dagster.config import DagsterConfig
from trilogyt.dagster.constants import SUFFIX

DEFAULT_DESCRIPTION: str = "No description provided"


@dataclass
class ModelInput:
    name: str
    path: Path

    @property
    def import_path(self) -> str:
        base = self.path.stem.replace("/", ".")
        return f"tests.integration.dagster.assets.optimization.{base}"


def generate_model_text(
    model_name: str,
    model_type: str,
    model_sql: str,
    dialect: Dialects,
    dependencies: list[ModelInput],
) -> str:
    template = Template(
        """
from dagster_duckdb import DuckDBResource
from dagster import asset
{% for dep in deps %}
from {{dep.import_path}} import {{dep.name}}
{% endfor %}

@asset(deps=[{% for dep in deps %}{{dep.name}}{% if not loop.last %}, {% endif %}{% endfor %}])
def {{model_name}}({{dialect.name | lower}}: DuckDBResource) -> None:
    with {{dialect.name | lower}}.get_connection() as conn:
        conn.execute(
           ''' {{model_sql}} '''
        )
    """
    )
    return template.render(
        model_name=model_name,
        model_type=model_type,
        model_sql=model_sql,
        dialect=dialect,
        deps=dependencies,
    )


def generate_entry_file(
    models: list[ModelInput],
    dialect: Dialects,  # config: DagsterConfig
):
    if dialect != Dialects.DUCK_DB:
        raise NotImplementedError(f"Unsupported dialect {dialect}")
    template = Template(
        """
from dagster import Definitions
from dagster_duckdb import DuckDBResource
{% for model in models %}
from {{model.importpath}} import {{model.name}}
{% endfor %}

defs = Definitions(
    assets=[{% for model in models %}{{model.name}}{% if not loop.last %}, {% endif %}{% endfor %}],
    resources={
        "{{dialect.value}}": DuckDBResource(
            database="", connection_config={"enable_external_access": False}
        )
    },
)"""
    )
    return template.render(
        models=models,
        dialect=Dialects,
    )


def generate_model(
    preql_body: str,
    preql_path: Path | None,
    dialect: Dialects,
    config: DagsterConfig,
    environment: Environment | None = None,
    clear_target_dir: bool = True,
) -> list[Path]:
    env: Environment = environment or Environment(
        working_path=preql_path.parent if preql_path else os.getcwd(),
        # namespace=config.namespace,
    )
    executor = Executor(
        dialect=dialect, engine=dialect.default_engine(), environment=env
    )
    executor.environment = enrich_environment(executor.environment)
    possible_dependencies = {}

    outputs: dict[str, str] = {}
    output_data: dict[str, Datasource] = {}
    try:
        _, statements = executor.environment.parse(preql_body)
    except Exception as e:
        raise SyntaxError(f"Unable to parse {preql_body}" + str(e))

    possible_dependencies = {
        d.identifier: d
        for d in executor.environment.datasources.values()
        if "." not in d.identifier
    }
    logger.info(Counter([type(c) for c in statements]))
    parsed = [z for z in statements if isinstance(z, PersistStatement)]

    logger.info("generating queries")
    logger.info(f"possible dependencies are {list(possible_dependencies.keys())}")
    logger.info([str(c) for c in executor.environment.materialized_concepts])

    pqueries = executor.generator.generate_queries(executor.environment, parsed)
    logger.info(f"got {len(pqueries)} queries")
    logger.info(Counter([type(c) for c in pqueries]))
    dependency_map = defaultdict(list)
    for _, query in enumerate(pqueries):
        depends_on: list[ModelInput] = []
        if isinstance(query, ProcessedQueryPersist):
            logger.info(f"Starting on {_}")
            target = query.output_to.address.location
            eligible = {k: v for k, v in possible_dependencies.items() if k != target}
            for cte in query.ctes:
                if isinstance(cte, UnionCTE):
                    continue
                # handle inlined datasources
                logger.info(f"checking cte {cte.name} with {eligible}")
                if cte.base_name_override in eligible:
                    if any(x.name == cte.base_name_override for x in depends_on):
                        continue
                    depends_on.append(
                        ModelInput(
                            cte.base_name_override,
                            config.get_asset_path(cte.base_name_override),
                        )
                    )
                for source in cte.source.datasources:
                    logger.info(source.identifier)
                    if not isinstance(source, Datasource):
                        continue

                    if source.identifier in eligible:
                        if any(x.name == source.identifier for x in depends_on):
                            continue
                        depends_on.append(
                            ModelInput(
                                source.identifier,
                                config.get_asset_path(source.identifier),
                            )
                        )
            # get our names to label the model
            key = query.output_to.address.location.split(".")[-1]
            outputs[key] = executor.generator.compile_statement(query)
            output_data[key] = query.datasource
            dependency_map[key] = depends_on
    logger.info("Writing queries to output files")
    logger.info(dependency_map)

    existing = defaultdict(set)
    should_exist = defaultdict(set)
    import_paths = []
    for key, value in outputs.items():

        output_path = config.get_asset_path(key)
        logger.info(f"writing {key} to {output_path} ")
        parent = str(output_path.parent)
        output_path.parent.mkdir(parents=True, exist_ok=True)
        for subf in output_path.parent.iterdir():
            if subf.is_file() and subf.name.endswith(SUFFIX):
                existing[parent].add(subf)
        should_exist[parent].add(output_path)
        with open(output_path, "w") as f:
            f.write(
                generate_model_text(
                    key,
                    "sql",
                    value,
                    dialect=dialect,
                    dependencies=dependency_map.get(key, []),
                )
            )
        import_paths.append(output_path)
    # config.config_path.parent.mkdir(parents=True, exist_ok=True)
    # with open(config.config_path, "w") as f:
    #     f.write(
    #         generate_entry_file(
    #             [ModelInput(name, path) for name, path in should_exist.items()],
    #             dialect,
    #             config,
    #         )
    #     )

    if clear_target_dir:
        for key, paths in existing.items():
            for path in paths:
                if path not in should_exist[key]:
                    logger.info("Removing old file: %s", path)
                    os.remove(path)
    return import_paths



import os
import tempfile
from pathlib import Path as PathlibPath

from trilogy.dialect.enums import Dialects

from trilogyt.constants import OPTIMIZATION_NAMESPACE, logger
from trilogyt.dagster.config import DagsterConfig
from trilogyt.dagster.generate import generate_model
from trilogyt.dagster.run import run_path
from trilogyt.scripts.native import OptimizationResult, native_wrapper


def dagster_handler(
    staging_path: PathlibPath,
    preql: PathlibPath,
    dagster_path: PathlibPath,
    dialect: Dialects,
    debug: bool,
    children: list[PathlibPath],
):
    logger.info("Optimizing trilogy files...")

    root = (
        native_wrapper(
            preql=preql,
            output_path=staging_path,
            dialect=dialect,
            debug=debug,
            run=False,
        )
        or {}
    )

    logger.info("Generating dagster models...")
    logger.info("clearing optimization path")
    opt = dagster_path / "models" / OPTIMIZATION_NAMESPACE

    if opt.exists():
        for item in opt.glob("*.sql"):
            logger.debug(f"Removing existing {item}")
            os.remove(item)
    import_paths = []
    for orig_file in children:
        file = staging_path / orig_file.name
        logger.info(
            f"Generating dagster model for {file} into dagster_path {dagster_path}"
        )
        # don't build hidden files
        if file.stem.startswith("_"):
            continue

        optimization: OptimizationResult | None = root.get(orig_file)

        if optimization:
            with open(optimization.path) as opt_file:
                opt_config = DagsterConfig(
                    root=PathlibPath(dagster_path), namespace=OPTIMIZATION_NAMESPACE
                )
                path = generate_model(
                    opt_file.read(),
                    optimization.path,
                    dialect=dialect,
                    config=opt_config,
                    clear_target_dir=False,
                    # environment = env  # type: ignore
                )
                import_paths += path
        config = DagsterConfig(root=PathlibPath(dagster_path), namespace=file.stem)
        with open(file) as f:
            path = generate_model(
                f.read(),
                file,
                dialect=dialect,
                config=config,
                # environment = env  # type: ignore
            )
            import_paths += path
    return import_paths


def dagster_wrapper(
    preql: PathlibPath,
    dagster_path: PathlibPath,
    dialect: Dialects,
    debug: bool,
    run: bool,
    staging_path: PathlibPath | None = None,
):
    imports: list[PathlibPath] = []
    if preql.is_file():
        config = DagsterConfig(root=dagster_path, namespace=preql.stem)
        with open(preql) as f:
            imports += generate_model(
                f.read(),
                preql,
                dialect=dialect,
                config=config,
                # environment = env  # type: ignore
            )
    else:
        children = list(preql.glob("*.preql"))
        if staging_path:
            imports += dagster_handler(
                staging_path, preql, dagster_path, dialect, debug, children
            )
        else:
            with tempfile.TemporaryDirectory() as tmpdirname:
                new_path = PathlibPath(tmpdirname)
                imports += dagster_handler(
                    new_path, preql, dagster_path, dialect, debug, children
                )
    # _ = generate_entry_file(imports, dialect)
    if run:
        print("Executing generated models")
        run_path(PathlibPath(dagster_path), imports=imports, dialect=dialect)
    return 0


def dagster_string_command_wrapper(
    preql: str, dagster_path: PathlibPath, dialect: Dialects, debug: bool, run: bool
):
    """handle a string command line input"""
    config = DagsterConfig(root=dagster_path, namespace="io")
    imports = generate_model(
        preql,
        dagster_path / "io.preql",
        dialect=dialect,
        config=config,
    )
    # _ = generate_entry_file(imports, dialect)
    if run:
        print("Executing generated models")
        run_path(PathlibPath(dagster_path), imports=imports, dialect=dialect)
    return 0



from trilogy.dialect.enums import Dialects  # noqa
from pathlib import Path as PathlibPath  # noqa
import os
from trilogyt.native.generate import generate_model
from trilogyt.native.run import run_path
from trilogyt.scripts.core import optimize_multiple, OptimizationResult
from trilogyt.constants import logger


def native_wrapper(
    preql: PathlibPath,
    output_path: PathlibPath,
    dialect: Dialects,
    debug: bool,
    run: bool,
) -> dict[PathlibPath, OptimizationResult]:
    logger.info(f"Running native wrapper with {preql} and {output_path}")

    existing = output_path.glob("**/*.preql")
    for item in existing:
        logger.debug(f"Removing existing {item}")
        os.remove(item)

    env_to_optimization = {}
    if preql.is_file():
        with open(preql) as f:
            generate_model(
                f.read(),
                preql,
                output_path=output_path,
                # environment = env  # type: ignore
            )
    else:
        # with multiple files, we can attempt to optimize dependency
        logger.info(f"checking path {preql}")
        children = [
            x for x in list(preql.glob("*.preql")) if not x.stem.startswith("_internal")
        ]
        logger.info(f"optimizing across {children}")
        env_to_optimization = optimize_multiple(
            preql, children, output_path, dialect=dialect
        )
        for file in children:
            with open(file) as f:
                with open(output_path / file.name, "w") as f2:
                    f2.write(f.read())
        for file in children:
            with open(file) as f:
                generate_model(
                    f.read(),
                    file,
                    output_path=output_path,
                    optimization=env_to_optimization.get(file, None),
                )

    if run:
        print("Executing generated models")
        run_path(output_path, dialect=dialect, env_to_optimization=env_to_optimization)
    return env_to_optimization


def native_string_command_wrapper(
    preql: str, output_path: PathlibPath, dialect: Dialects, debug: bool, run: bool
):
    """handle a string command line input"""
    generate_model(
        preql_body=preql,
        # this doesn't exist, but will after we run
        preql_path=output_path / "io.preql",
        output_path=output_path,
        # environment = env  # type: ignore
    )
    if run:
        print("Executing generated models")
        run_path(
            output_path,
            dialect=dialect,
        )
    return 0



from pathlib import Path

from networkx import DiGraph, topological_sort
from trilogy import Dialects, Environment

from trilogyt.constants import logger
from trilogyt.scripts.core import OptimizationResult


def generate_execution_order(edges):
    graph = DiGraph()
    for edge in edges:
        graph.add_edge(*edge)
    return topological_sort(graph)


def run_path(
    path: Path,
    dialect: Dialects,
    env_to_optimization: dict[Path, OptimizationResult] | None = None,
):
    # initialize
    files = path.glob("*.preql")
    edges = []
    executor = dialect.default_executor()
    for x in files:
        try:
            env = Environment(working_path=path)
            executor.environment = env
            executor.parse_file(x)
            for _, imp_list in env.imports.items():
                for imp in imp_list:
                    edges.append((imp.path, x))

        except Exception as e:
            logger.error(f" Error executing {x} {e}")
            raise e

    if env_to_optimization:
        logger.info("Have optimization scripts, running builds for datasources first")
        opt_build_scripts = set()
        for _, v in env_to_optimization.items():
            opt_build_scripts.add(v.path)
        for script in opt_build_scripts:
            env = Environment(working_path=path)
            executor.environment = env
            executor.execute_file(script)
    sorted_files: list[Path] = generate_execution_order(edges)
    for file in sorted_files:
        env = Environment(working_path=path)
        executor.environment = env
        if not file.suffix == "preql":
            file = file.with_suffix(".preql")
        executor.execute_file(path / file, non_interactive=True)



import os
from collections import Counter, defaultdict
from pathlib import Path

from jinja2 import Template
from trilogy import Environment, Executor
from trilogy.core.models import (
    Address,
    Datasource,
    PersistStatement,
    ProcessedQuery,
    ProcessedQueryPersist,
    UnionCTE,
)
from trilogy.dialect.enums import Dialects
from yaml import dump, safe_load

from trilogyt.constants import logger
from trilogyt.core import enrich_environment
from trilogyt.dbt.config import DBTConfig

DEFAULT_DESCRIPTION: str = "No description provided"


def generate_model_text(model_name: str, model_type: str, model_sql: str) -> str:
    template = Template(
        """
    {{ model_type }} "{{ model_name }}" {
        {{ model_sql }}
    }
    """
    )
    return template.render(
        model_name=model_name, model_type=model_type, model_sql=model_sql
    )


def generate_model(
    preql_body: str,
    preql_path: Path | None,
    dialect: Dialects,
    config: DBTConfig,
    environment: Environment | None = None,
    clear_target_dir: bool = True,
):
    env: Environment = environment or Environment(
        working_path=preql_path.parent if preql_path else os.getcwd(),
        # namespace=config.namespace,
    )
    executor = Executor(
        dialect=dialect, engine=dialect.default_engine(), environment=env
    )
    executor.environment = enrich_environment(executor.environment)
    possible_dependencies = {}

    outputs: dict[str, str] = {}
    output_data: dict[str, Datasource] = {}
    try:
        _, statements = executor.environment.parse(preql_body)
    except Exception as e:
        raise SyntaxError(f"Unable to parse {preql_body}" + str(e))

    possible_dependencies = {
        d.identifier: d for d in executor.environment.datasources.values()
    }
    logger.info(Counter([type(c) for c in statements]))
    parsed = [z for z in statements if isinstance(z, PersistStatement)]

    logger.info("generating queries")
    logger.info(f"possible dependencies are {list(possible_dependencies.keys())}")
    logger.info([str(c) for c in executor.environment.materialized_concepts])

    pqueries = executor.generator.generate_queries(executor.environment, parsed)
    logger.info(f"got {len(pqueries)} queries")
    logger.info(Counter([type(c) for c in pqueries]))
    for _, query in enumerate(pqueries):
        if isinstance(query, ProcessedQueryPersist):
            logger.info(f"Starting on {_}")
            target = query.output_to.address.location
            eligible = {k: v for k, v in possible_dependencies.items() if k != target}
            for cte in query.ctes:
                # handle inlined datasources
                logger.info(f"checking cte {cte.name} with {eligible}")
                if isinstance(cte, UnionCTE):
                    continue
                if cte.base_name_override in eligible:
                    cte.base_name_override = (
                        f"{{{{ ref('{cte.base_name_override}_gen_model') }}}}"
                    )
                for source in cte.source.datasources:
                    logger.info(source.identifier)
                    if not isinstance(source, Datasource):
                        continue

                    if source.identifier in eligible:
                        if isinstance(source.address, Address):
                            source.address.location = (
                                f"{{{{ ref('{source.identifier}_gen_model') }}}}"
                            )
                        elif isinstance(source.address, str):
                            source.address = (
                                f"{{{{ ref('{source.identifier}_gen_model') }}}}"
                            )
            base = ProcessedQuery(
                output_columns=query.output_columns,
                ctes=query.ctes,
                base=query.base,
                joins=query.joins,
                grain=query.grain,
                hidden_columns=query.hidden_columns,
                limit=query.limit,
                where_clause=query.where_clause,
                order_by=query.order_by,
            )
            # get our names to label the model
            outputs[query.output_to.address.location.split(".")[-1]] = (
                executor.generator.compile_statement(base)
            )
            output_data[query.output_to.address.location.split(".")[-1]] = (
                query.datasource
            )
    logger.info("Writing queries to output files")

    existing = defaultdict(set)
    should_exist = defaultdict(set)
    for key, value in outputs.items():

        output_path = config.get_model_path(key)
        logger.info(f"writing {key} to {output_path} ")
        parent = str(output_path.parent)
        output_path.parent.mkdir(parents=True, exist_ok=True)
        for subf in output_path.parent.iterdir():
            if subf.is_file() and subf.name.endswith("gen_model.sql"):
                existing[parent].add(subf)
        should_exist[parent].add(output_path)
        with open(output_path, "w") as f:
            if preql_path:
                f.write(
                    f"-- Generated from preql source: {preql_path.stem}\n-- Do not edit manually\n"
                )
            # set materialization here
            # TODO: make configurable
            f.write("{{ config(materialized='table') }}\n")
            f.write(value)

    config.config_path.parent.mkdir(parents=True, exist_ok=True)
    if config.config_path.exists():
        with open(config.config_path, "r") as f:
            loaded = safe_load(f.read())
    else:
        loaded = {}
    loaded["version"] = 2
    loaded["models"] = loaded.get("models", [])
    models = []
    for key, value in outputs.items():
        ds = output_data[key]
        columns = [
            {
                "name": c.alias,
                "description": (
                    c.concept.metadata.description or DEFAULT_DESCRIPTION
                    if c.concept.metadata
                    else DEFAULT_DESCRIPTION
                ),
                "tests": ["unique"] if [c] == ds.grain.components else [],
            }
            for c in ds.columns
        ]
        nobject = {
            "name": f"{key}_gen_model",
            "description": "Automatically generated model from trilogy",
            "columns": columns,
        }
        # loaded["models"] = [x for x in loaded["models"] if x["name"] != nobject["name"]]
        models.append(nobject)

    with open(config.config_path, "w") as f:
        loaded["models"] = models
        f.write(dump(loaded))

    if clear_target_dir:
        for key, paths in existing.items():
            for path in paths:
                if path not in should_exist[key]:
                    logger.info("Removing old file: %s", path)
                    os.remove(path)



import importlib.util
import os
from pathlib import Path
from typing import Any

from trilogy import Dialects

from trilogyt.constants import logger


def import_asset_from_file(filepath: Path) -> Any:
    """
    Dynamically imports an asset from a file based on the filename.

    Args:
        filepath (Path): The path to the Python file containing the asset.
    """
    # Get the file name without extension
    filename = os.path.splitext(os.path.basename(filepath))[0]

    # Load the module from the file path
    spec = importlib.util.spec_from_file_location(filename, filepath)
    if spec is None or spec.loader is None:
        raise ImportError(f"Cannot load module from {filepath}")

    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    # Access the asset with the same name as the file
    target_object = filename.rsplit("_", 2)[0]
    if not hasattr(module, target_object):
        raise AttributeError(f"Asset '{target_object}' not found in {filepath}")

    return getattr(module, target_object)


def run_path(path: Path, imports: list[Path], dialect: Dialects):
    from dagster import materialize

    selection = []
    for file in set(imports):
        logger.info(f"found file: {file}")
        selection.append(import_asset_from_file(file))
    if not selection:
        raise SyntaxError("No assets found in path")
    resources = {}
    if dialect == Dialects.DUCK_DB:
        from dagster_duckdb import DuckDBResource

        resources = {
            "duck_db": DuckDBResource(
                database="dagster.db",  # required
                # tests wll error
                connection_config={"enable_external_access": False},
            )
        }
    else:
        raise NotImplementedError(f"Unsupported dialect: {dialect}")

    result = materialize(assets=selection, resources=resources, selection=selection)
    print(f"Job result: {result}")



from trilogy.dialect.enums import Dialects  # noqa
from pathlib import Path as PathlibPath  # noqa
from trilogyt.constants import OPTIMIZATION_NAMESPACE  # noqa
from trilogyt.dbt.generate import generate_model  # noqa
from trilogyt.dbt.run import run_path  # noqa
from trilogyt.dbt.config import DBTConfig  # noqa
from trilogyt.scripts.native import native_wrapper, OptimizationResult
from trilogyt.constants import logger
import os
import tempfile


def dbt_handler(
    staging_path: PathlibPath,
    preql: PathlibPath,
    dbt_path: PathlibPath,
    dialect: Dialects,
    debug: bool,
    run: bool,
    children: list[PathlibPath],
):
    logger.info("Optimizing trilogy files...")

    root = (
        native_wrapper(
            preql=preql,
            output_path=staging_path,
            dialect=dialect,
            debug=debug,
            run=False,
        )
        or {}
    )

    logger.info("Generating dbt models...")
    logger.info("clearing optimization path")
    opt = dbt_path / "models" / OPTIMIZATION_NAMESPACE

    if opt.exists():
        for item in opt.glob("*.sql"):
            logger.debug(f"Removing existing {item}")
            os.remove(item)
    for orig_file in children:
        file = staging_path / orig_file.name
        logger.info(f"Generating dbt model for {file} into dbt_path {dbt_path}")
        # don't build hidden files
        if file.stem.startswith("_"):
            continue

        optimization: OptimizationResult | None = root.get(orig_file)

        if optimization:
            with open(optimization.path) as opt_file:
                opt_config = DBTConfig(
                    root=PathlibPath(dbt_path), namespace=OPTIMIZATION_NAMESPACE
                )
                generate_model(
                    opt_file.read(),
                    optimization.path,
                    dialect=dialect,
                    config=opt_config,
                    clear_target_dir=False,
                    # environment = env  # type: ignore
                )
        config = DBTConfig(root=PathlibPath(dbt_path), namespace=file.stem)
        with open(file) as f:
            generate_model(
                f.read(),
                file,
                dialect=dialect,
                config=config,
                # environment = env  # type: ignore
            )


def dbt_wrapper(
    preql: PathlibPath,
    dbt_path: PathlibPath,
    dialect: Dialects,
    debug: bool,
    run: bool,
    staging_path: PathlibPath | None = None,
):
    if preql.is_file():
        config = DBTConfig(root=dbt_path, namespace=preql.stem)
        with open(preql) as f:
            generate_model(
                f.read(),
                preql,
                dialect=dialect,
                config=config,
                # environment = env  # type: ignore
            )
    else:
        children = list(preql.glob("*.preql"))
        if staging_path:
            dbt_handler(staging_path, preql, dbt_path, dialect, debug, run, children)
        else:
            with tempfile.TemporaryDirectory() as tmpdirname:
                new_path = PathlibPath(tmpdirname)
                dbt_handler(new_path, preql, dbt_path, dialect, debug, run, children)

    if run:
        print("Executing generated models")
        run_path(PathlibPath(dbt_path))
    return 0


def dbt_string_command_wrapper(
    preql: str, dbt_path: PathlibPath, dialect: Dialects, debug: bool, run: bool
):
    """handle a string command line input"""
    config = DBTConfig(root=dbt_path, namespace="io")
    generate_model(
        preql,
        dbt_path / "io.preql",
        dialect=dialect,
        config=config,
    )
    if run:
        print("Executing generated models")
        run_path(PathlibPath(dbt_path))
    return 0
